import typing

from google.cloud.datastore import Query
from gumo.core.injector import injector

from gumo.datastore.infrastructure import EntityKeyMapper
from gumo.datastore.infrastructure.repository import DatastoreRepositoryMixin


class DatastoreRepositoryMixinForTest(DatastoreRepositoryMixin):
    KIND = None
    KINDS = None

    @classmethod
    def get_entity_key_mapper(cls) -> EntityKeyMapper:
        if cls._entity_key_mapper is None:
            cls._entity_key_mapper: EntityKeyMapper = injector.get(EntityKeyMapper)

        return cls._entity_key_mapper

    @classmethod
    def cleanup_entities(cls):
        if cls.KIND is not None:
            cls.cleanup_entities_of_kind(kind=cls.KIND)

        if cls.KINDS is not None and isinstance(cls.KINDS, typing.Iterable):
            for kind in cls.KINDS:
                cls.cleanup_entities_of_kind(kind=kind)

    @classmethod
    def count_entities(cls) -> int:
        if cls.KIND is None:
            raise RuntimeError('KIND must be present.')

        return cls.count_entities_of_kind(kind=cls.KIND)

    @classmethod
    def cleanup_entities_of_kind(cls, kind: str):
        client = cls.get_datastore_client()
        query = client.query(kind=kind)
        query.keys_only()
        while True:
            keys = [entity.key for entity in query.fetch(limit=500)]
            if len(keys) == 0:
                break
            client.delete_multi(keys=keys)

    @classmethod
    def count_entities_of_kind(cls, kind: str) -> int:
        query = cls.get_datastore_client().query(kind=kind)
        query.keys_only()
        return len(list(query.fetch()))

    @classmethod
    def fetch_kinds(cls):
        query = cls.get_datastore_client().query(kind="__kind__")
        query.keys_only()

        kinds = [entity.key.id_or_name for entity in query.fetch()]
        return kinds

    @classmethod
    def cleanup_all_entities(cls):
        for kind in cls.fetch_kinds():
            cls.cleanup_entities_of_kind(kind=kind)

    @classmethod
    def fetch_keys_of_kind(cls, kind: str):
        entity_key_mapper = cls.get_entity_key_mapper()
        client = cls.get_datastore_client()
        query = client.query(kind=kind)
        query.keys_only()
        for entity in query.fetch():
            yield entity_key_mapper.to_entity_key(datastore_key=entity.key)

    @classmethod
    def fetch_keys_of_query(cls, query: Query):
        entity_key_mapper = cls.get_entity_key_mapper()
        query.keys_only()
        for entity in query.fetch():
            yield entity_key_mapper.to_entity_key(datastore_key=entity.key)



from gumo.datastore.infrastructure.alias import DatastoreEntity
from gumo.datastore.infrastructure.alias import DatastoreKey
from gumo.datastore.infrastructure.alias import DatastoreClient
from gumo.datastore.infrastructure.alias import DatastoreQuery
from gumo.datastore.infrastructure.repository import DatastoreRepositoryMixin
from gumo.datastore.infrastructure.repository import datastore_transaction
from gumo.datastore.infrastructure.entity_key_mapper import EntityKeyMapper
from gumo.datastore.infrastructure.mapper import DatastoreMapperMixin
from gumo.datastore.infrastructure.model import DataModel

__all__ = [
    DatastoreRepositoryMixin.__name__,

    DatastoreEntity.__name__,
    DatastoreKey.__name__,
    DatastoreClient.__name__,
    DatastoreQuery.__name__,

    datastore_transaction.__name__,
    EntityKeyMapper.__name__,
    DatastoreMapperMixin.__name__,
    DataModel.__name__,
]



import setuptools


name = 'gumo-datastore'
version = '0.1.10'
description = 'Gumo Datastore Library'
dependencies = [
    'gumo-core >= 0.1.3',
    'google-cloud-datastore >= 1.7.1, >= 1.8.0',
]

with open("README.md", "r") as fh:
    long_description = fh.read()

packages = [
    package for package in setuptools.find_packages()
    if package.startswith('gumo')
]

namespaces = ['gumo']

setuptools.setup(
    name=name,
    version=version,
    author="Gumo Project Team",
    author_email="gumo-py@googlegroups.com",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gumo-py/gumo-datastore",
    packages=packages,
    namespaces=namespaces,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=dependencies,
)



import typing
from contextlib import contextmanager
from injector import singleton

from gumo.core.injector import injector
from gumo.core import EntityKey
from gumo.core import NoneKey
from gumo.datastore.infrastructure.configuration import DatastoreConfiguration
from gumo.datastore.infrastructure.entity_key_mapper import EntityKeyMapper
from gumo.datastore.infrastructure.alias import DatastoreClient
from gumo.datastore.infrastructure.alias import DatastoreEntity
from gumo.datastore.infrastructure.alias import DatastoreKey
from gumo.datastore.infrastructure.alias import DatastoreQuery


class DatastoreRepositoryMixin:
    _datastore_client = None
    _entity_key_mapper = None

    DatastoreEntity = DatastoreEntity

    @classmethod
    def get_datastore_client(cls) -> DatastoreClient:
        if cls._datastore_client is None:
            configuration: DatastoreConfiguration = injector.get(DatastoreConfiguration, scope=singleton)
            cls._datastore_client = configuration.client

        return cls._datastore_client

    @property
    def datastore_client(self) -> DatastoreClient:
        if self._datastore_client is None:
            self.get_datastore_client()

        return self._datastore_client

    @property
    def entity_key_mapper(self) -> EntityKeyMapper:
        if self._entity_key_mapper is None:
            self._entity_key_mapper: EntityKeyMapper = injector.get(EntityKeyMapper)

        return self._entity_key_mapper

    def to_entity_key(self, datastore_key: typing.Optional[DatastoreKey]) -> EntityKey:
        return self.entity_key_mapper.to_entity_key(datastore_key=datastore_key)

    def to_datastore_key(self, entity_key: typing.Union[EntityKey, NoneKey, None]) -> typing.Optional[DatastoreKey]:
        return self.entity_key_mapper.to_datastore_key(entity_key=entity_key)

    def _fetch_page(
        self, query: DatastoreQuery, limit: int, start_cursor: typing.Optional[bytes]
    ) -> typing.Tuple[typing.List[DatastoreEntity], typing.Optional[str], bool]:
        query_iterator = query.fetch(limit=limit, start_cursor=start_cursor)
        pages = query_iterator.pages
        page = next(pages)
        entities = list(page)

        next_cursor = (
            query_iterator.next_page_token.decode("utf-8")
            if query_iterator.next_page_token
            else None
        )
        if next_cursor:
            has_more = len(list(query.fetch(limit=1, start_cursor=next_cursor))) > 0
        else:
            has_more = False

        if not has_more:
            next_cursor = None

        return entities, next_cursor, has_more


@contextmanager
def datastore_transaction():
    datastore_client: DatastoreClient = injector.get(DatastoreConfiguration).client

    with datastore_client.transaction():
        yield
